/*
 *  Copyright (C) 2008, Pedro Carlos da Silva Lara
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version. 
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.   
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package ec.zp;
import java.math.BigInteger;
import java.util.HashMap;

public class Point {

    /**
     * @param x primeira coordenada
     * @param y segunda coordenada
     * @param inftyPoint true caso seja um ponto no infinito, 
     *        false caso contrario
     * @param E curva que este ponto pertence
     */
    public Point( BigInteger x, BigInteger y, boolean inftyPoint, Curve E ) {
        this.x = x;
        this.y = y;
        this.inftyPoint = inftyPoint;
        this.E = E;
    }
    /**
     * Construtor para pontos que nao sejam no infinito
     * 
     * @param x primeira coordenada
     * @param y segunda coordenada
     * @param E curva que este ponto pertence
     */
    public Point( BigInteger x, BigInteger y, Curve E ) {
        this.x = x;
        this.y = y;
        this.E = E;
        inftyPoint = false;
    }
    
    /**
     * Costrutor por copia
     * 
     * @param P sera copiado
     */
    
    public Point( Point P ) {
        this.x = P.getX();
        this.y = P.getY();
        this.inftyPoint = P.isInftyPoint();
        this.E = P.getCurve();
    }
    /**
     * Costroi um ponto no infinito
     * 
     * @param E curva que este ponto pertencer
     */
    public Point( Curve E ) {
    	this.x = new BigInteger("0");
    	this.y = new BigInteger("0");
    	this.inftyPoint = true;
    	this.E = E;
    }
    public byte[] toArrayByte() {
        byte [] retbytes = new byte[150];
        
        if( this.isInftyPoint() )
            retbytes[0] = (byte) 0x01;
        else
            retbytes[0] = (byte) 0x00;
        
        byte[] xbytes = x.toByteArray();
        
        int i;
        for(i = 1; i <= xbytes.length; i++)
            retbytes[i] = xbytes[i - 1];
        
        byte[] ybytes = y.toByteArray();
        
        for(int j = 0; j < ybytes.length; j++){
            retbytes[i] = xbytes[j];
            i++;
        }
        return retbytes;
    }
    /**
     * 
     * @return O ponto <b>this</b> formatado
     */
    public String getString() {
        if ( inftyPoint )
            return "infty point";
        else
            return "("+this.x.toString()+ ", "+this.y.toString()+")";
    }
    
    /**
     * 
     * @return A primeira coordenada
     */
    
    public BigInteger getX(){
        return x;
    }
    
    /**
     * 
     * @return A segunda coordenada
     */
    public BigInteger getY(){
        return y;
    }
    /**
     * 
     * @return true se  um ponto no infinito, false caso contrrio
     */
    public boolean isInftyPoint(){
        return inftyPoint;
    }
    
    /**
     * 
     * @return A curva que o ponto <b>this</b> pertence
     */
    public Curve getCurve() {
        return E;
    }
    
    /**
     * Compara o ponto <b>this</b> com o ponto <b>P</b>
     * 
     * @param P ponto que ser comparado
     * @return true se so iguais, false caso contrrio
     */
    
    public boolean equals( Point P ) {
        return ( this.x.equals( P.getX() )           &&
                 this.y.equals( P.getY() )           &&
                 this.inftyPoint == P.isInftyPoint() &&
                 this.E.equals( P.getCurve() ) );
    }
    /**
     * Calcula o inverso aditivo do Ponto <b>this</b>
     * 
     * @return -this
     */
    
    public Point negate() {
        return new Point( x, E.getP().subtract( y ), E);
    }
    
    /**
     * Multiplica o escalar <b>K</b> pelo ponto <b>this</b>
     * 
     * @param K Inteiro que multiplicar o ponto <b>this</b>
     * @return R onde <b>R = K * this</b>
     */
    
    public Point multiply( BigInteger K ) {
        Point Q   = new Point( BigInteger.ZERO, BigInteger.ZERO, true, this.E );
        Point P   = this;
        
        int numBitsOfK = K.bitLength();
        
        for ( int i = numBitsOfK - 1; i >= 0; i-- ) {
            Q = Q.add( Q );
        	if( K.testBit( i ) )
                Q = Q.add( P );
        }
        
        return Q;
        
    }
    
    public Point multiply_NAF( BigInteger K ) {
        int [] k = Utils.NAF( K, (byte) 2 );
        Point Q   = new Point( BigInteger.ZERO, BigInteger.ZERO, true, this.E );
        Point P   = this;
        
        int l = 0;
        
        // Correcao do tamanho do vetor k
        
        if( k[k.length - 1] == 0)
        	l = k.length  - 1;
        else
        	l = k.length;
        
        for( int i = l - 1; i >= 0; i-- ) {
            Q = Q.doublePoint();
            if( k[i] == 1 ) {
                Q = Q.add( P );
            } else if( k[i] == -1 ) {
            	Q = Q.add( P.negate() );
            }
            /*System.out.println("Q  = "+ Q.getString());
            System.out.println("P  = "+ P.getString());
            System.out.println("ki = "+ k[i]);
            System.out.println("i  = "+ i+"\n\n");*/
       }
        
        return Q;
    }
    
    public Point multiply_NAF2w( BigInteger K, byte w ) {
    	//long t0 = System.currentTimeMillis();
        HashMap<Integer, Point> h = Utils.PreComputingNAF2w( this, w);
        //long t1 = System.currentTimeMillis();
       // System.out.println("Pre2^"+w+": time = "+ (t1 - t0));
        
        Point Q   = new Point( BigInteger.ZERO, BigInteger.ZERO, true, this.E );
        int [] k = Utils.NAF( K, w );
        
        for ( int  i = k.length - 1 ; i >= 0; i-- ) {
        	Q = Q.doublePoint();
        	if( k[i] != 0 ) {
        		if( k[i] > 0 ) {
        			Q = Q.add( h.get( k[i] ) );
        		} else {
        			Q = Q.add( h.get( -k[i] ).negate() );
        		}
        	}
        }
        return Q;
   }
    
    /**
     * Operao de soma entre pontos  
     * 
     * @param P Ponto que ser somado
     * @return <b>R</b>, onde <b>R = this + P</b>
     */
    
    public Point add( Point P ) {
    	if( inftyPoint )
            return P;
    	
        if( P.isInftyPoint() )
            return this;
        
    	if( this.equals( P ) )
            return doublePoint();
    	
        if( this.equals( P.negate() ) )
        	return new Point( E );
        
        BigInteger t    = new BigInteger("0");
        BigInteger xRet = new BigInteger("0");
        BigInteger yRet = new BigInteger("0");
        
        BigInteger aux  = (P.getX().subtract(this.x)).mod( E.getP() );
        
        BigInteger tmp = aux.modInverse( E.getP() );
        
        t = ((P.getY().subtract(this.y)).multiply(tmp)).mod(E.getP());
        xRet = ((( t.multiply( t ) ).subtract( this.x ))
                .subtract( P.getX() ) ).mod( E.getP() );
        yRet = ( (t.multiply( this.x.subtract( xRet ) ) )
                .subtract( this.y ) ).mod( E.getP() ); 
        return new Point( xRet, yRet, this.E );
    }
    
    /**
     * Operao de dobra de pontos
     * 
     * @return <b>R</b>, onde <b>R = 2 * this</b>
     */
    
    private Point doublePoint() {

    	if( !( this.y.gcd( E.getP()) ).equals( BigInteger.ONE ) )
            return new Point( BigInteger.ZERO, BigInteger.ZERO, true, this.E );
        
        BigInteger t    = new BigInteger("0");
        BigInteger xRet = new BigInteger("0");
        BigInteger yRet = new BigInteger("0");
     
        BigInteger tmp  = ((new BigInteger("2")).multiply(this.y))
                .modInverse(E.getP());
        
        t = ((((new BigInteger("3")).multiply(this.x.multiply(this.x)))
                .add(E.getA())).multiply(tmp)).mod(E.getP());
        
        xRet = ((t.multiply(t)).subtract( (new BigInteger("2")
                .multiply( this.x )))).mod(E.getP());
        
        yRet = ((t.multiply( this.x.subtract(xRet)))
                .subtract(this.y)).mod(E.getP()); 
        
        return new Point( xRet, yRet, false, this.E );
    }
    
    private BigInteger x;
    private BigInteger y;
    private boolean inftyPoint;
    private Curve E;
}